# Copyright 2024 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Color 4.8.8 quantum error correction code implementation.

This module contains the implementation of the Color 4.8.8 quantum error correction code,
including patch definitions, gate implementations, and compilation language options.
"""

from enum import Enum

from pecos.slr.qeclib.color488.color488_class import Color488Patch


class Language(Enum):
    """Language options to compile SLR to."""

    QASM = 0
    QIR = 1
    QIRBC = 2
