# Copyright 2025 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""QASM regression tests for Steane non-flagged Z measurement."""

from collections.abc import Callable

from pecos.slr import CReg, QReg
from pecos.slr.qeclib.steane.meas.measure_z import NoFlagMeasureZ


def test_MeasureX(compare_qasm: Callable[..., None]) -> None:
    """Test Steane non-flagged Z measurement QASM regression."""
    q = QReg("q_test", 7)
    a = QReg("a_test", 1)
    out = CReg("out_test", 1)

    block = NoFlagMeasureZ(q, a, out)
    compare_qasm(block)
