"""Error models for quantum simulations.

This package provides various error models for simulating noise in quantum circuits,
including depolarizing, Pauli, and circuit-based error models.
"""

# Copyright 2018 National Technology & Engineering Solutions of Sandia, LLC (NTESS). Under the terms of Contract
# DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from pecos.error_models.class_errors_circuit import ErrorCircuits
from pecos.error_models.old.depolar_gen import DepolarModel
from pecos.error_models.old.gatewise_gen import GatewiseModel
from pecos.error_models.old.xerror_gen import XModel
from pecos.error_models.old.xzerror_gen import XZModel
from pecos.error_models.old.zerror_gen import ZModel

__all__ = [
    "DepolarModel",
    "ErrorCircuits",
    "GatewiseModel",
    "XModel",
    "XZModel",
    "ZModel",
]
