"""Single-qubit Hadamard gate implementations.

This module provides Hadamard gate implementations for creating superposition
states and performing basis transformations between computational and
Hadamard bases in quantum error correction protocols.
"""

# Copyright 2024 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from pecos.slr.qeclib.qubit.qgate_base import QGate


class H(QGate):
    """Hadamard gate.

    This gate creates superposition states and transforms between X and Z bases.
    """
